import configparser
import json
import logging
import time
import unittest
from opcua import Client
import slapos.client

class EndToEndTestCase(unittest.TestCase):
  @classmethod
  def setUpClass(cls):
    configp = configparser.ConfigParser()
    configp.read('${slapos-client.cfg:output}')
    args = type("empty_args", (), {})()
    conf = slapos.client.ClientConfig(args, configp)
    local = slapos.client.init(conf, logging.getLogger(__name__))
    cls.slap = local['slap']
    cls._supply = staticmethod(local['supply'])
    cls._request = staticmethod(local['request'])
    cls.product = staticmethod(local['product'])
    cls._requested = {}
    cls._supplied = {}

  @classmethod
  def tearDownClass(cls):
    # remove instances
    for args, kw in cls._requested.values():
      kw['state']  = 'destroyed'
      cls._request(*args, **kw)
    # remove software releases
    for k,v in cls._supplied.items():
      print("Remove %s from %s" %(k, v))
      cls._supply(k, v, 'destroyed')

  @classmethod
  def request(cls, *args, **kw):
    instance_name = args[1]
    cls._requested[instance_name] = (args, kw)
    partition = cls._request(*args, **kw)
    return cls.unwrapConnectionDict(partition.getConnectionParameterDict())

  @classmethod
  def supply(cls, software_release, computer_id, state):
    cls._supplied[software_release] = computer_id
    cls._supply(software_release, computer_id, state)

  @staticmethod
  def unwrapConnectionDict(connection_dict):
    try:
      connection_json = connection_dict['_']
    except KeyError:
      return connection_dict
    return json.loads(connection_json)

  @classmethod
  def getInstanceInfos(cls, instance_name):
    # adapated from cli/info
    infos = cls.slap.registerOpenOrder().getInformation(instance_name)
    class Infos:
      def __init__(self, **kw):
        self.__dict__.update(kw)
    connection_dict = {
      e['connection_key'] : e['connection_value']
      for e in infos._connection_dict
    }
    return Infos(
      software_url = infos._software_release_url,
      software_type = infos._source_reference,
      shared = bool(infos._root_slave),
      requested_state = infos._requested_state,
      parameter_dict = infos._parameter_dict,
      connection_dict = cls.unwrapConnectionDict(connection_dict),
      news = infos._news,
    )

  @classmethod
  def getInstanceNews(cls, instance_name):
    try:
      news = cls.slap.registerOpenOrder().getInformation(instance_name)._news
    except Exception:
      return ()
    return news['instance']

  @classmethod
  def getInstanceStatus(cls, instance_name):
    # adapated from cli/info
    status = 0b00
    for e in cls.getInstanceNews(instance_name):
      text = e.get('text', '')
      if text.startswith('#access'):
        status |= 0b01
      elif text.startswith('#error'):
        status |= 0b10
      if status == 0b11:
        break
    return ('none', 'green', 'red', 'orange')[status]

  @classmethod
  def waitUntilGreen(cls, instance_name, timeout=80):
    t0 = time.time()
    while (status := cls.getInstanceStatus(instance_name)) != 'green':
      msg = 'Instance %s status is still %s' % (instance_name, status)
      print(msg)
      if (time.time() - t0) > 60 * timeout:
        raise TimeoutError(msg)
      time.sleep(60)


class BeremizTest(EndToEndTestCase):

  # XXX: move outside runner at Test Suite ?
  #computer_id = "COMP-3523" # Ubunto 20.04
  #computer_id = "COMP-4036" # Debian 10
  computer_id = "COMP-4039" # Debian 11
  #computer_id = "COMP-4042"  # Ubunto 22.04
 
  coupler_release = "https://lab.nexedi.com/nexedi/slapos/raw/master/software/osie-coupler/software-dev.cfg"
  runtime_release = "https://lab.nexedi.com/nexedi/slapos/raw/master/software/beremiz-runtime/software.cfg"

  @classmethod   
  def setUpClass(cls):
    super().setUpClass()

    # supply needed software releases
    print("Supply software releases at %s" %cls.computer_id)
    cls.supply(cls.coupler_release, cls.computer_id, state="available")
    cls.supply(cls.runtime_release, cls.computer_id, state="available")

    # XXX: rather than sleep add check to wait for SRs compilation end
    print("Sleep for 4h.")
    time.sleep(4*3600)

    # supply / request coupler
    instance_name = time.strftime('e2e-test-coupler-%Y-%B-%d-%H:%M:%S')
    parameter_dict = {"mode":1,
                      "network_interface": "192.168.0.0"}
    cls.request(cls.coupler_release,
                instance_name,
                partition_parameter_kw=parameter_dict)
    cls.waitUntilGreen(instance_name, 180)
    connection_dict = cls.getInstanceInfos(instance_name).connection_dict
    print(connection_dict)
    cls.coupler_url_ipv6 = connection_dict.get("url-ipv6")

    # supply / request beremiz-runtime
    instance_name = time.strftime('e2e-test-beremiz-runtime-%Y-%B-%d-%H:%M:%S')
    parameter_dict = {"runtime_plc_url": "https://lab.nexedi.com/nexedi/osie/raw/master/Beremiz/beremiz_test_opc_ua/bin/beremiz_test_opc_ua.tgz"}
    cls.request(cls.runtime_release,
                instance_name,
                partition_parameter_kw=parameter_dict)
    cls.waitUntilGreen(instance_name, 180)
    connection_dict = cls.getInstanceInfos(instance_name).connection_dict
    print(connection_dict)

  def test_plc_increment_run(self):
    NUMBER_OF_CHECKS = 100
    TIMEOUT = 2
    OPC_UA_IDENTIFIER = "ns=1;s=i2c0.relay0"

    # give it some time for services(runtime & coupler) to warm up, connect and run
    time.sleep(30)

    # connect to a session at OPC-UA server
    client = Client(self.coupler_url_ipv6)

    # for now this is the only test thus we start it without a wrapper
    test_count = 1
    test_failures = 0
    expected_failures = 0
    try:
      client.connect()
      root = client.get_root_node()
      children_list = root.get_children()
      var = client.get_node(OPC_UA_IDENTIFIER)
      for i in range (0, NUMBER_OF_CHECKS):
        i2c0_relay0_before = var.get_value()
        print("\ni2c0_relay0 (before) = ", i2c0_relay0_before)
        print("Sleep for %s seconds ..." %TIMEOUT)
        time.sleep(TIMEOUT)
        i2c0_relay0_after = var.get_value()
        print("i2c0_relay0 (after) = ", i2c0_relay0_after)
        # for the wait timeout runtime should have increased the value
        if (i2c0_relay0_after <= i2c0_relay0_before):
          # counter should have been increased, mark failure
          test_failures += 1
    finally:
      client.disconnect()
    # no failures
    self.assertEqual(test_failures, 0)
